{ Pas2Po - Extract Strings out of an GNU Pascal source and write .po file
  Copyright (C) 2001 Eike Lange <eike@gnu.de>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
 
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
 
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
}

program pas2po;

uses GPC, GPCUtil;

type
  PStringStack = ^TStringStack;
  TStringStack = record
    Next: PStringStack;
    Entry: TString;      { the string to translate }
    IsCString: Boolean;  { true, if CString }
    AddInfo: TString;    { Filename, Line Number }
  end;

var
  GetAllStrings, Verbose: Boolean value false;
  WriteLocation: Boolean value true;
  FileNames: array [1..255] of TString;
  NumInFiles, Index: Integer;
  InFileName, OutFileName: TString value '-';
  StringStack: PStringStack;

{ Output a help message and exit. }
procedure Help;
begin
  WriteLn ('Pas2Po 0.2 - Copyright (C) 2001 Eike Lange <eike@gnu.de>');
  WriteLn ('This program is free software under the GNU General Public License.');
  WriteLn ('Usage: ', ParamStr (0), ' [Options] [input.pas [...]]');
  WriteLn ('Valid options are:');
  WriteLn (' --help     -h              output this help text and exit');
  WriteLn (' --all      -a              extract ALL strings');
  WriteLn (' --verbose  -v              tell verbosely what is happening');
  WriteLn (' --output   -o <file>       write output to <file> instead of stdout');
  WriteLn (' --no-location              do not write comments about location');
  Halt (0)
end;

{ Use GNU GetOpt to parse the command line. }
procedure ParseOptions;
const
  LongOptions: array [1..5] of OptionType =
      (('help', NoArgument, nil, 'h'),
       ('all', NoArgument, nil, 'a'),
       ('verbose', NoArgument, nil, 'v'),
       ('output', RequiredArgument, nil, 'o'),
       ('no-location', NoArgument, nil, #1));
var
  LongIndex, Index: Integer;
  Ch: Char;
begin
  if ParamCount = 0 then
    begin
      Help;
      Halt(0)
    end;
  LongIndex := -1;
  repeat
    Ch := GetOptLong ('', LongOptions, LongIndex, false);
    case Ch of
      'h': Help;
      'v': Verbose := true;
      'a': GetAllStrings := true;
      'o': OutFileName := OptionArgument;
       #1: WriteLocation := false;
    end
  until Ch = EndOfOptions;
  NumInFiles := 0;
  for Index := FirstNonOption to ParamCount do
    begin
      Inc (NumInfiles);
      FileNames[NumInFiles] := ParamStr (Index)
    end;
end;

{ initializes StringStack }
procedure SStackInit;
begin
  new (StringStack);
  StringStack^.next := nil;
  StringStack^.Entry := '';
  StringStack^.IsCString := false;
  StringStack^.AddInfo := ''
end;

{ destroys StringStack }
procedure SStackDestroy;
var 
  p, e: PStringStack;
begin
  p := StringStack;
  while p^.Next <> nil do
    begin
      e := p^.Next;
      p^.Next := e^.Next;
      Dispose (e)
    end;
  Dispose (p)
end;

{ adds a String to stack, does not, if stack contains string
  already }
procedure SStackAdd (sEntry: TString; sCString: Boolean; 
  sAddInfo: TString);
var
  p: PStringStack;
  Found: Boolean;  { true if string was found in stack }
begin
  Found := sEntry = '';
  p := StringStack;
  while (p^.Next <> nil) and (not Found) do
    begin
      Found := p^.Entry = sEntry;
      if Found then 
        begin
          p^.AddInfo := p^.AddInfo + ' ' + sAddInfo;
          if Verbose then
            WriteLn (StdErr, 'Duplicate Entry `', sEntry, 
              ''' found. Add location.')
        end;
      p := p^.next
    end;
  if not Found then
    begin
      { p^.next is nil, so its an empty element }
      if Verbose then
        WriteLn (StdErr, 'Add Entry `', sEntry, '''.');
      p^.Entry := sEntry;
      p^.IsCString := sCString;
      p^.AddInfo := sAddInfo;
      new (p^.next);
      p^.next^.entry := '';
      p^.next^.next := nil
    end
end;

{ prints out Stack }
procedure SStackPrint (var f: Text);
var
  p: PStringStack;
begin
  p := StringStack;
  while p^.Entry <> '' do
    begin
      WriteLn (f);
      if WriteLocation then
        WriteLn (f, '#' + p^.AddInfo);
      WriteLn (f, 'msgid "', p^.Entry, '"');
      WriteLn (f, 'msgstr ""');
      p := p^.Next
    end
end;

{ Finds out the beginning of a string }
function StringBeginPos (aLine: TString): Integer;
var
  s: CharSet value ['"', ''''];
begin
  StringBeginPos := CharPos (s, aLine)
end;

{ reads a pascal string }
procedure GetPasString (var aFile: Text; var aLine: TString; 
  var aLineNo: Integer; aPos: Integer; aStackIt: Boolean);
const
  DoublePQuote = '''''';
  SinglePQuote = '''';
var
  DoubleQuotePos, SingleQuotePos: Integer;
  TheString: TString value '';
  FoundEnd: Boolean value false;
  LineSave: Integer;
begin
  aLine := SubStr (aLine, aPos + 1);
  LineSave := aLineNo;
  while not FoundEnd do
    begin
      DoubleQuotePos := Pos (DoublePQuote, aLine);
      SingleQuotePos := Pos (SinglePQuote, aLine);
      if  (SingleQuotePos <> 0) and 
          ((DoubleQuotePos = 0) or (DoubleQuotePos > SingleQuotePos)) then
        begin
          TheString := TheString + SubStr (aLine, 1, SingleQuotePos - 1);
          aLine := SubStr (aLine, SingleQuotePos + 1);
          FoundEnd := true
        end
      else if (DoubleQuotePos = SingleQuotePos) and 
              (DoubleQuotePos <> 0) then
        begin
          TheString := TheString + SubStr (aLine, 1, DoubleQuotePos);
          aLine := SubStr (aLine, DoubleQuotePos + 2)
        end
      else if SingleQuotePos = 0 then
        begin
          TheString := TheString + aLine + NewLine;
          ReadLn (aFile, aLine);
          Inc (aLineNo)
        end
    end;
  if aStackIt then
    SStackAdd (TheString, false, InFileName + ':' + Int2Str (LineSave) )
end;

{ Reads a C String }
procedure GetCString (var aFile: Text; var aLine: TString; 
  var aLineNo: Integer; aPos: Integer; aStackIt: Boolean);
const
  EscapeCQuote = '\"';
  SingleCQuote = '"';
var
  EscapeQuotePos, SingleQuotePos: Integer;
  TheString: TString value '';
  FoundEnd: Boolean value false;
  LineSave: Integer;
begin
  aLine := SubStr (aLine, aPos + 1);
  LineSave := aLineNo;
  while not FoundEnd do
    begin
      EscapeQuotePos := Pos (EscapeCQuote, aLine);
      SingleQuotePos := Pos (SingleCQuote, aLine);
      if  (SingleQuotePos <> 0) and 
          ((EscapeQuotePos = 0) or (EscapeQuotePos > SingleQuotePos)) then
        begin
          TheString := TheString + SubStr (aLine, 1, SingleQuotePos - 1);
          aLine := SubStr (aLine, SingleQuotePos + 1);
          FoundEnd := true
        end
      else if (EscapeQuotePos < SingleQuotePos) and 
              (EscapeQuotePos <> 0) then
        begin
          TheString := TheString + SubStr (aLine, 1, EscapeQuotePos + 1);
          aLine := SubStr (aLine, EscapeQuotePos + 2)
        end
      else if SingleQuotePos = 0 then
        begin
          TheString := TheString + aLine + '\n';
          ReadLn (aFile, aLine);
          Inc (aLineNo)
        end
    end;
  if aStackIt then
    SStackAdd (TheString, true, InFileName + ':' + Int2Str (LineSave) )
end;

{ Read a String and put it onto the Stack }
procedure ReadString (var aFile: Text; var aLine: TString; 
  var aLineNo: Integer; StringPos: Integer; StackIt: Boolean);
begin
  if aLine[StringPos] = '''' then  { its a GNU Pascal String }
    GetPasString (aFile, aLine, aLineNo, StringPos, StackIt)
  else  { its a C String }
    GetCString (aFile, aLine, aLineNo, StringPos, StackIt)
end;

{ Finds out the beginning of a comment }
function CommentBeginPos (aLine: TString): Integer;
var
  Pos1, Pos2, Pos3: Integer;
begin
  Pos1 := Pos ('{',  aLine);
  Pos2 := Pos ('(*', aLine);
  Pos3 := Pos ('//', aLine);
  if (Pos1 = 0) and (Pos2 = 0) and (Pos3 = 0) then
    CommentBeginPos := 0
  else
    begin
      if Pos1 = 0 then Pos1 := High (Integer);
      if Pos2 = 0 then Pos2 := High (Integer);
      if Pos3 = 0 then Pos3 := High (Integer);
      if (Pos1 < Pos2) and (Pos1 < Pos3) then 
        CommentBeginPos := Pos1
      else if (Pos2 < Pos3) then
        CommentBeginPos := Pos2
      else
        CommentBeginPos := Pos3
    end
end;

{ Reads to the end of comment }
procedure FindCommentEnd (var aFile:Text; var aLine: TString; 
  var aLineNo:Integer; CommentPos: Integer);
var
  CommentType: Char;
  CommentCounter: Integer value 1;
  CommentLength, CmtEnd, CmtBegin: Integer value 0;
  CommentBeginString, CommentEndString: String(2);
begin
  CommentType := aLine[CommentPos];
  if CommentType = '/' then 
    begin
      aLine := '';
      CommentCounter := 0
    end;
  if CommentType = '{' then
    begin
      CommentLength := 1;
      CommentBeginString := '{';
      CommentEndString := '}'
    end
  else if CommentType = '(' then
    begin
      CommentLength := 2;
      CommentBeginString := '(*';
      CommentEndString := '*)'
    end;
  if CommentCounter > 0 then
    begin
      aLine := SubStr (aLine, CommentPos+1);
      while CommentCounter <> 0 do
        begin
          CmtEnd := Pos (CommentEndString, aLine);
          CmtBegin := Pos (CommentBeginString, aLine);
          if (CmtEnd = 0) and (CmtBegin = 0) then
            begin
              ReadLn (aFile, aLine);
              Inc (aLineNo);
            end
          else if (CmtEnd <> 0) and ( (CmtEnd < CmtBegin) or 
            (CmtBegin = 0)) then
            begin
              Dec (CommentCounter);
              aLine := SubStr (aLine, CmtEnd + CommentLength)
            end
          else if (CmtBegin <> 0) and ( (CmtBegin < CmtEnd) or 
            (CmtEnd = 0)) then
            begin
              Inc (CommentCounter);
              aLine := SubStr (aLine, CmtBegin + CommentLength)
            end
        end
    end
end;

{ Finds out a "GetText" in aLine }
function GettextBeginPos (aLine: TString): Integer;
const
  Maxx = High (Integer);
  BeforeString = [' ', '+', '(', ',', '='];
  AfterString = [' ', '('];
var
  SearchStrings: array [1..6] of String(12) value
    ('GetText',   'GetTextC', 
     'DGetText',  'DGetTextC',
     'DcGetText', 'DcGetTextC');
  Pos: array [1..6] of Integer;
  i: Byte;
  Min: Integer value Maxx;
  GtLength: Integer value 0;
begin
  Pos[1] := PosCase (SearchStrings[1], aLine);
  Pos[2] := PosCase (SearchStrings[2], aLine);
  Pos[3] := PosCase (SearchStrings[3], aLine);
  Pos[4] := PosCase (SearchStrings[4], aLine);
  Pos[5] := PosCase (SearchStrings[5], aLine);
  Pos[6] := PosCase (SearchStrings[6], aLine);
  for i := 1 to 6 do
    if (Pos[i] < Min) and (Pos[i] <> 0) then 
      begin
        Min := Pos[i];
        GtLength := Length (SearchStrings[i])
      end;
  if Min = Maxx then
    GettextBeginPos := 0
  else if Min = 1 then
    begin
      if (GtLength + Min - 1) = Length (aLine) then
        GettextBeginPos := Min
      else if aLine[Min + GtLength] in AfterString then
        GettextBeginPos := Min
      else
        GettextBeginPos := 0
    end
  else if aLine[Min -1] in BeforeString then
    begin
      if (GtLength + Min - 1) = Length (aLine) then
        GettextBeginPos := Min
      else if aLine[Min + GtLength] in AfterString then
        GettextBeginPos := Min
      else
        GettextBeginPos := 0
    end
  else
    GettextBeginPos := 0
end;

{ Read the string from GetText(C)|DGetText(C)|DcGetText(C) }
procedure ReadGettext (var aFile: Text; var aLine: TString; 
  var aLineNo: Integer; StringPos: Integer);
const
  StringBegin  = ['''', '"'];
  LeftBracket  = ['('];
  RightBracket = [')'];
  Comma        = [','];
var
  StringBeginPos, LeftBracketPos, RightBracketPos, 
    CommaPos: Integer;
begin
  aLine := SubStr (aLine, StringPos);
  if (PosCase ('GetText',  aLine) = 1) or 
     (PosCase ('GetTextC', aLine) = 1) then
    begin
      LeftBracketPos  := CharPos (LeftBracket, aLine);
      { Multiline Gettext? }
      while LeftBracketPos = 0 do
        begin
          ReadLn (aFile, aLine);
          Inc (aLineNo);
          LeftBracketPos  := CharPos (LeftBracket, aLine)
        end;
      aLine := SubStr (aLine, LeftBracketPos + 1)
    end
  else if (PosCase ('DGetText',   aLine) = 1) or 
          (PosCase ('DGetTextC',  aLine) = 1) or
          (PosCase ('DcGetText',  aLine) = 1) or 
          (PosCase ('DcGetTextC', aLine) = 1) then
    begin
      LeftBracketPos  := CharPos (LeftBracket, aLine);
      { Multiline DGettext? }
      while LeftBracketPos = 0 do
        begin
          ReadLn (aFile, aLine);
          Inc (aLineNo);
          LeftBracketPos  := CharPos (LeftBracket, aLine)
        end;
      aLine := SubStr (aLine, LeftBracketPos);
      { find the next comma }
      CommaPos := CharPos (Comma, aLine);
      while CommaPos = 0 do
        begin
          ReadLn (aFile, aLine);
          Inc (aLineNo);
          CommaPos := CharPos (Comma, aLine)
        end;
      aLine := SubStr (aLine, CommaPos + 1);
    end;
  { the next argument is to translate }
  RightBracketPos := CharPos (RightBracket, aLine);
  StringBeginPos  := CharPos (StringBegin, aLine);
  while (StringBeginPos = 0) and (RightBracketPos = 0) do
    begin
      ReadLn (aFile, aLine);
      Inc (aLineNo);
      RightBracketPos := CharPos (RightBracket, aLine);
      StringBeginPos  := CharPos (StringBegin, aLine);
    end;
  if (StringBeginPos <> 0) and 
    ((StringBeginPos < RightBracketPos) or (RightBracketPos = 0)) then
    begin
      if aLine[StringBeginPos] = '''' then
        GetPasString (aFile, aLine, aLineNo, StringBeginPos, true)
      else
        GetCString (aFile, aLine, aLineNo, StringBeginPos, true)
    end
end;

{ reads the file }
procedure ReadFile (aFileName: TString);
var
  f: Text;
  SrcLine: TString;
  LineNo, CommentPos, StringPos, StringDel: Integer value 0;
begin
  Assign (f, aFileName);
  Reset (f);
  while not Eof (f) do
    begin
      ReadLn (f, SrcLine);
      Inc (LineNo);
      repeat
        if GetAllStrings then
          StringPos := StringBeginPos (SrcLine)
        else
          { if GetText is in a String - discard it }
          begin
            StringPos := GettextBeginPos (SrcLine);
            StringDel := StringBeginPos (SrcLine);
            while ((StringDel < StringPos) and (StringDel <> 0)) or
                  ((StringDel <> 0) and (StringPos = 0)) do
              begin
                ReadString (f, SrcLine, LineNo, StringDel, false);
                StringPos := GettextBeginPos (SrcLine);
                StringDel := StringBeginPos (SrcLine)
              end
          end;
        CommentPos := CommentBeginPos (SrcLine);
        if (CommentPos <> 0) and 
           ( (CommentPos < StringPos) or (StringPos=0)) then
          FindCommentEnd (f, SrcLine, LineNo, CommentPos)
        else if StringPos <> 0 then
          begin
            if GetAllStrings then
              ReadString (f, SrcLine, LineNo, StringPos, true)
            else
              { if GetText is in a String - discard it }
              begin
                if StringPos < StringBeginPos (SrcLine) then
                  ReadGettext (f, SrcLine, LineNo, StringPos)
                else
                  ReadString (f, SrcLine, LineNo, StringPos, false)
              end;
          end
      until (CommentPos = 0) {and (StringPos = 0)}
    end;
  Close (f)
end;

{ some comments, uses Global: InFileName }
procedure OutHeader (var f: Text);
begin
  WriteLn (f, '# This file was created by pas2po with ''', InFileName, '''.');
  WriteLn (f, '# Please change this file manually.');
  WriteLn (f, '# SOME DESCRIPTIVE TITLE.');
  WriteLn (f, '# Copyright (C) YEAR Free Software Foundation, Inc.');
  WriteLn (f, '# FIRST AUTHOR <EMAIL@ADDRESS>, YEAR.');
  WriteLn (f, '#');
  WriteLn (f, '#, fuzzy')
end;

{ Write tho PO-Header. }
procedure OutPoHeader (var f: Text);
const
  aNullS: array [Boolean] of String (1) = ('', '0');
var
  t: TimeStamp;
  s: String (100);
begin
  GetTimestamp(t);
  WriteLn (f, 'msgid ""');
  WriteLn (f, 'msgstr ""');
  WriteLn (f, '"Project-Id-Version: PACKAGE VERSION\n"');
  s := '"POT-Creation-Date: ' + Int2Str(t.Year) + '-';
  s := s + aNullS[t.Month<10] + Int2Str(t.Month) + '-';
  s := s + aNullS[t.Day<10] + Int2Str(t.Day) + ' ';
  s := s + aNullS[t.Hour<10] + Int2Str(t.Hour) + ':';
  {@@ FIXME: Zone-Information actually not available in GNU Pascal }
  s := s + aNullS[t.Minute<10] + Int2Str(t.Minute) + '+ZONE\n"';
  WriteLn (f, s);
  WriteLn (f, '"PO-Revision-Date: YEAR-MO-DA HO:MI+ZONE\n"');
  WriteLn (f, '"Last-Translator: FULL NAME <EMAIL@ADDRESS>\n"');
  WriteLn (f, '"Language-Team: LANGUAGE <LL@li.org>\n"');
  WriteLn (f, '"MIME-Version: 1.0\n"');
  WriteLn (f, '"Content-Type: text/plain; charset=CHARSET\n"');
  WriteLn (f, '"Content-Transfer-Encoding: ENCODING\n"')
end;

{ writes PO-Information to file }
procedure OutputPOFile (aFileName: TString);
var
  f: Text;
begin
  if Verbose then Write (StdErr, 'Write output to ', aFileName, ': '); 
  Assign (f, aFileName);
  Rewrite (f);
  if Verbose then Write (StdErr, 'Header, '); 
  OutHeader (f);
  if Verbose then Write (StdErr, 'POHeader '); 
  OutPOHeader (f);
  if Verbose then Write (StdErr, 'and Data...'); 
  SStackPrint (f);
  if Verbose then WriteLn (StdErr, '..done.'); 
  Close (f)
end;

begin
  ParseOptions;
  SStackInit;
  if NumInFiles = 0 then 
    ReadFile (InFileName)
  else
    for Index := 1 to NumInFiles do
      begin
        InFileName := FileNames[Index];
        ReadFile (InFileName)
      end;
  OutputPOFile (OutFileName);
  SStackDestroy
end.
